package org.freecompany.brimstone.module;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedList;
import org.freecompany.util.core.Pair;
import org.freecompany.util.multicaster.Notification;

import static java.util.Collections.enumeration;
import static java.util.Collections.list;
import static org.freecompany.util.multicaster.NotificationMulticaster.logging;

/**
 * The delegating class loader maintains a list of delegates to which looading commands
 * will be propagated. This list is always traversed in order of insertion.
 */
public class DelegatingClassLoader extends URLClassLoader {

	private final Collection< ClassLoader> delegates = new LinkedList< ClassLoader>();
	private final ThreadTracker tracker = new ThreadTracker();

	protected Notification notifier = logging();

	/**
	 * Creates an delegating class loader with no parent and no
	 * source {@link URL}s.
	 */
	public DelegatingClassLoader() {
		this(( ClassLoader) null);
	}

	/**
	 * Creates an delegating class loader with no source {@link URL}s
	 * which uses the provided parent loader.
	 */
	public DelegatingClassLoader( final ClassLoader parent) {
		this( new URL[ 0], parent);
	}

	/**
	 * Creates an delegating class loader with no parent and with the
	 * specified source {@link URL}s.
	 */
	public DelegatingClassLoader( final URL[] urls) {
		this( urls, null);
	}

	/**
	 * Creates an delegating class loader with the specified source {@link URL}s
	 * and which uses the provided parent loader.
	 */
	public DelegatingClassLoader( final URL[] urls, final ClassLoader parent) {
		super( urls, parent);
	}

	/**
	 * Exposes the ability to add locations to this class loader at runtime. This
	 * implementation does not promote the existing {@link URLClassLoader#addURL} method
	 * so as to avoid interfering with normal operation.
	 *
	 * @param url the url to add to the list of locations searched for class definitions.
	 */
	public void addUrl( final URL url) {
		notifier.debug( "Adding URL ''{0}'' to class loader.", url);
		super.addURL( url);
	}

	/**
	 * Adds a class loader to the list of delegates.	Please note that the provided
	 * {@link ClassLoader} <b> must not</b> have this delegating class loader set as
	 * it's parent <b>unless</b> it is an instance of {@link InvertedClassLoader} or
	 * can otherwise make use of the methods in this loader that take a source as a
	 * parameter.
	 *
	 * @param loader the {@link ClassLoader} to add to the delegate list.
	 */
	public void addDelegate( final ClassLoader loader) {
		notifier.debug( "Adding delegate loader ''{0}''.", loader);
		delegates.add( loader);
	}

	/**
	 * Loads a class using the contained set of delegates.  This method conforms to the
	 * standard delegation model, checking the parent first if one exists.  If not, each
	 * child class loader is tried, in order, until one returns positively.  State is
	 * tracked via a thread local variable.
	 *
	 * @param name the name of the class to load.
	 * @return the {@link Class} definition found by this loader.
	 * @throws ClassNotFoundException if the class can not be found.
	 */
	@Override
	public Class loadClass( final String name, final boolean resolve) throws ClassNotFoundException {
		if ( !searching( name)) {
			notifier.debug( "{0}.loadClass( {1})", shortName( this.getClass()), name);
			try {
				return super.loadClass( name, resolve);
			} catch ( ClassNotFoundException e) {
				for ( ClassLoader child : delegates) {
					try {
						push( new Pair< ClassLoader, String>( child, name));
						try {
							return child.loadClass( name);
						} finally {
							pop();
						}
					} catch ( ClassNotFoundException e2) {
						assert true;
					}
				}
			}
			notifier.debug( "Failed to find ''{0}''.", name);
		}
		throw new ClassNotFoundException( name);
	}

	/**
	 * Gets a set of resources using the contained set of delegates. The parent, if
	 * one exists, and each child are utilized to build a complete set of resources
	 * for the given name. State is tracked via a thread local variable.
	 *
	 * @param name the name of the resource to list.
	 * @return the {@link Enumeration< URL>} of resources found by this class loader.
	 */
	@Override
	public Enumeration< URL> getResources( final String name) throws IOException {
		final Collection< URL> resources = new LinkedList< URL>();
		if ( !searching( name)) {
			notifier.debug( "{0}.getResources( {1})", shortName( this.getClass()), name);
			
			Enumeration< URL> enumeration = super.getResources( name);
			if ( enumeration.hasMoreElements()) return enumeration;

			for ( ClassLoader child : delegates) {
				if ( skip( child, name)) notifier.debug( "Skipping source child ''{0}'' for resource ''{1}''.", child, name);
				else {
					push( new Pair< ClassLoader, String>( child, name));
					try {
						resources.addAll( list( child.getResources( name)));
					} finally {
						pop();
					}
				}
			}
			if ( resources.isEmpty()) notifier.debug( "Failed to find resources for ''{0}''.", name);
		}
		return enumeration( resources);
	}

	/**
	 * Gets a resource using the contained set of delegates.  This method conforms to the
	 * standard delegation model, checking the parent first if one exists.  If not, each
	 * child class loader is tried, in order, until one returns positively.  State is
	 * tracked via a thread local variable.
	 *
	 * @param name the name of the resource to get.
	 * @return the {@link URL} of the resource found by this class loader or
	 * <code>null</code> if no resource was found.
	 */
	@Override
	public URL getResource( final String name) {
		if ( !searching( name)) {
			notifier.debug( "{0}.getResource( {1})", shortName( this.getClass()), name);
			URL url;
			url = super.getResource( name);
			if ( url != null) return url;

			for ( ClassLoader child : delegates) {
				if ( skip( child, name)) notifier.debug( "Skipping source child ''{0}'' for resource ''{1}''.", child, name);
				else {
					push( new Pair< ClassLoader, String>( child, name));
					try {
						url = child.getResource( name);
					} finally {
						pop();
					}

					if ( url != null) return url;
				}
			}
			notifier.debug( "Failed to find resource ''{0}''.", name);
		}
		return null;
	}

	/**
	 * Strips extra characters from a fully qualified class name, leaving only
	 * the name within the package.
	 *
	 * @param clazz the {@link Class} for which to construct a short name.
	 * @return the class name within its package.
	 */
	protected final static String shortName( final Class clazz) {
		return clazz.getName().replaceAll( ".*[.]", "");
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append( shortName( getClass())).append( "[ ");
		for ( URL url : getURLs()) builder.append( url.toString().replaceAll( ".*/", "")).append( ", ");
		builder.setLength( builder.length() - 2);
		return builder.append( "]").toString();
	}


	/**
	 * Pushes a new {@link Pair} onto the thread local stack.
	 *
	 * @param pair the pair of elements describing the latest loading event.
	 */
	protected void push( final Pair< ClassLoader, String> pair) {
		tracker.get().push( pair);
	}

	/**
	 * Pops the {@link Pair} on the top of the stack.
	 *
	 * @return the top of the stack if it is not empty.
	 * @throws EmptyStackException if there are no entries on the stack.
	 */
	protected Pair< ClassLoader, String> pop() {
		try {
			return tracker.get().pop();
		} finally {
			if ( tracker.get().isEmpty()) tracker.remove();
		}
	}

	/**
	 * Checks to see if the provided name is currently in the process of being
	 * loaded. This indicates that this thread is still within a {@link ClassLoader#loadClass( String)}
	 * call.
	 *
	 * @param name the current class being loaded.
	 */
	protected boolean searching( final String name) {
		if ( tracker.get().isEmpty()) return false;
		return tracker.get().peek().two().equals( name);
	}

	/**
	 * Checks the thread local state to determine whether the provided child needs
	 * to be skipped.
	 *
	 * @param child child class loader to test.
	 * @param name the name of the resource being loaded.
	 * @return <code>boolean</code> indicating whether this child should be skipped.
	 */
	protected boolean skip( final ClassLoader child, final String name) {
		if ( tracker.get().isEmpty()) return false;
		final Pair< ClassLoader, String> pair = tracker.get().peek();
		return pair.one() == child && pair.two().equals( name);
		//for ( Pair< ClassLoader, String> pair : tracker.get()) if ( pair.one() == child && pair.two().equals( name)) return true;
		//return false;
	}

	/**
	 * Dumps a string representation of the current threads local stack.
	 *
	 * @return {$link String} representing the current stack state.
	 */
	protected final String dump() {
		final StringBuilder builder = new StringBuilder( "Stack[ ");
		builder.append( tracker.get().size()).append( "]\n");
		for ( Pair< ClassLoader, String> pair : tracker.get()) {
			builder.append( pair.one()).append( " -> ").append( pair.two()).append( "\n");
		}
		return builder.toString();
	}
}
