package org.freecompany.brimstone.module;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Stack;

import static java.util.Collections.enumeration;

/**
 * Inverts the basic class loader delegation model to allow
 * the same resource to exist in multiple locations within
 * a single tree of class loaders.
 */
public class InvertedClassLoader extends URLClassLoader {

	/**
	 * Creates an inverted class loader with no parent and no
	 * source {@link URL}s.
	 */
	public InvertedClassLoader() {
		this(( ClassLoader) null);
	}

	/**
	 * Creates an inverted class loader with no source {@link URL}s
	 * which uses the provided parent loader.
	 */
	public InvertedClassLoader( final ClassLoader parent) {
		this( new URL[ 0], parent);
	}

	/**
	 * Creates an inverted class loader with no parent and with the
	 * specified source {@link URL}s.
	 */
	public InvertedClassLoader( final URL[] urls) {
		this( urls, null);
	}

	/**
	 * Creates an inverted class loader with the specified source {@link URL}s
	 * and which uses the provided parent loader.
	 */
	public InvertedClassLoader( final URL[] urls, final ClassLoader parent) {
		super( urls, parent);
	}

	/**
	 * In this class loader the standard {@link URLClassLoader#addURL} method
	 * is promoted to allow modifying the list of {@link URL}s at runtime.
	 */
	@Override
	public void addURL( final URL url) {
		super.addURL( url);
	}

	/**
	 * Reverses the ordering of the normal {@link URLClassLoader#loadClass} method to
	 * first search this loader and to delegate only if the class is
	 * not found. This loader is aware of the {@link DelegatingClassLoader} and
	 * will interact with it appropriately to eliminate class loading loops.
	 */
	@Override
	public Class< ?> loadClass( final String name) throws ClassNotFoundException {
		try {
			return findClass( name);
		} catch ( ClassNotFoundException e) {
			if ( getParent() != null) return getParent().loadClass( name);
			throw new ClassNotFoundException( name);
		}
	}

	/**
	 * Reverses the ordering of the normal {@link URLClassLoader#getResource} method to
	 * first search this loader and to delegate only if the resource is
	 * not found. This loader is aware of the {@link DelegatingClassLoader} and
	 * will interact with it appropriately to eliminate class loading loops.
	 */
	@Override
	public URL getResource( final String name) {
		final URL resource = findResource( name);
		if ( resource != null) return resource;
		return getParent() == null ? null : getParent().getResource( name);
	}

	/**
	 * Reverses the ordering of the normal {@link URLClassLoader#getResources} method to
	 * first search this loader and to delegate only if no resource can be found. This
	 * loader is aware of the {@link DelegatingClassLoader} and will interact with it
	 * appropriately to eliminate class loading loops.
	 */
	@Override
	public Enumeration< URL> getResources( final String name) throws IOException {
		final Enumeration< URL> resources = findResources( name);
		if ( resources.hasMoreElements()) return resources;
		return getParent() == null ? enumeration( Collections.< URL>emptySet()) : getParent().getResources( name);
	}
}
