package org.freecompany.brimstone.module;

import org.freecompany.brimstone.urn.URNStreamHandlerFactory;
import org.freecompany.util.multicaster.Notification;
import org.freecompany.util.xml.editor.XmlEditor;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URLConnection;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import static java.net.URL.setURLStreamHandlerFactory;
import static java.net.URLConnection.getFileNameMap;
import static java.net.URLConnection.setFileNameMap;
import static java.net.URLConnection.setContentHandlerFactory;
import static org.freecompany.brimstone.cache.Util.loadToCache;
import static org.freecompany.brimstone.core.Util.installCache;
import static org.freecompany.brimstone.core.Util.isBrimstoneUri;
import static org.freecompany.util.multicaster.NotificationMulticaster.logging;

/**
 * A module is a collection of components and dependencies as well as relevant meta data.  A
 * module is serialized as an XML document instance of the schema
 * <a href='http://www.freecompany.org/namespace/brimstone/module'>http://www.freecompany.org/namespace/brimstone/module</a>
 * which contains this information.
 * <p/>
 * A component consists of a URI describing a resource, either as a URL or a Brimstone specific
 * URI that can be resolved to a particular resource, and an optional component name.  The component
 * name can be used to reference a discrete component and defaults to the file name as interpreted from the
 * URI or URL specification.  Components are exposed by name in order to decouple the particular
 * implementation version from it's final use.  A component URI may reference another module definition.  In
 * this case, the module behaves as if the component module's component and dependency (as well as other) entries
 * existed within the local module.  Therefore accessing the component or dependency lists will return a superset
 * of the components of both this module and it's component modules.  Furthermore, alias lookups will resolve within
 * both the current module and any component modules.
 * <p/>
 * A dependency consists of a URI matching the convention of a component declaration and has no associated name.
 * <p/>
 * All operations within this class assume that the provided descriptor has been validated against the schema. If this
 * is not the case and the document is not valid, the behavior is undefined. In particular {@link ClassCastException}s
 * might be thrown from various methods that attempt to manipulate the document.
 */
public class Module {

	public static final String NAMESPACE = "http://www.freecompany.org/namespace/brimstone/module";
	public static final String CONTENT_TYPE = "application/brimstone-module";
	public static final String[] EXTRACT_NAME = { ".*?(\\w+)((?:-\\w+)*?)(?:-[\\w|.]+)?([.]\\w+)", "$1$2$3"};

	protected static final Notification notifier = logging();
	protected static Map< String, Module> cache = Collections.synchronizedMap( new HashMap< String, Module>());

	/**
	 * Initializes the JVM static factories for connection handling.  This static block eliminates
	 * multiple initializations on this class object, but if this class is loaded in parallel in another
	 * class loader multiple invocations could be possible.  So, we catch the {@link Error} that is
	 * generated in that case.
	 * <p/>
	 * Note: this means that it is also possible that if this class is loaded after the application
	 * has set a content handler that does not support the {@link Module} type or a stream handler that
	 * does not recognize the "brimstone" URN namespace it will fail silently and cause strange behavior,
	 * in particular excetions relating to unsupported protocols and {@link URL#getContent} calls which
	 * do not successfully return {@link Module} instances when they should.
	 */
	static {
		notifier.debug( "Initializing module system.");
		installCache();
		setFileNameMap( new ContentTypeMap( getFileNameMap()));

		try {
			setContentHandlerFactory( new HandlerFactory());
		} catch ( NoClassDefFoundError e) {
			throw e;
		} catch ( Error e) {
			notifier.warning( "Attempt to reinitialize content handler.", e);
		}

		try {
			setURLStreamHandlerFactory( new URNStreamHandlerFactory());
		} catch ( NoClassDefFoundError e) {
			throw e;
		} catch ( Error e) {
			notifier.warning( "Attempt to reinitialize stream handler.", e);
		}
	}

	protected final XmlEditor editor = new XmlEditor();
	protected List< Module> resolved;
	protected URL base;

	/**
	 * Creates a {$link Module} by loading the XML content referenced by the provided connection.
	 * This constructor behaves the same as calling {$link #Module( URLConnection, boolean)} with
	 * the second parameter <code>false</code>.
	 *
	 * @param connection the connection from which to load the descriptor.
	 * @throws IOException if the content cannot be loaded.
	 * @throws SAXException if the content is not a valid XML document.
	 */
	public Module( final URLConnection connection) throws IOException, SAXException {
		this( connection.getURL());
	}

	/**
	 * Creates a {$link Module} by loading the XML content referenced by the provided connection
	 * and optionally resolves all references.
	 *
	 * @param connection the connection from which to load the descriptor.
	 * @param resolve flag indicating whether references should be initially resolved.
	 * @throws IOException if the content cannot be loaded.
	 * @throws SAXException if the content is not a valid XML document.
	 */
	public Module( final URLConnection connection, final boolean resolve) throws IOException, SAXException {
		this( connection.getURL(), resolve);
	}

	/**
	 * Creates a new module from the given {@link URL} without resolving modules. This constructor
	 * behaves the same as calling {$link #Module( URL, boolean)} with the second parameter <code>false</code>.
	 *
	 * @param url {@link URL} the url of the module definition to load.
	 * @throws IOException if the content cannot be loaded.
	 * @throws SAXException if the content is not a valid XML document.
	 */
	public Module( final URL url) throws IOException, SAXException {
		this( url, false);
	}

	/**
	 * Creates a new module from the given URL, optionally resolving component modules.  If the resolve flag is
	 * <code>true</code>, any component {@link URI} that points to a module definition will be preloaded and indexed
	 * for use in resolving aliases.  When resolved in this fashion, calls to {@link #getComponent( String alias)} will
	 * automatically fallthrough to included component modules as well.
	 * <p/>
	 * Note that this constructor does not simply pass the resolve flag to the next constructor.  This is because the
	 * base {@link URL} must be set before attempting to resolve the module.
	 *
	 * @param url {@link URL} of the module definition to load.
	 * @param resolve flag indicating whether to resolve component modules.
	 */
	public Module( final URL url, final boolean resolve) throws IOException, SAXException {
		editor.read( url.openStream());
		editor.startPrefixMapping( NAMESPACE, "module");
		this.base = url;
		if ( resolve) resolve();
	}

	/**
	 * Retrieves the {$link URL} of this module descriptor. This {$link URL} can
	 * be used for building more complex {$link URL} strings that can resolve aliases
	 * within the this descriptor through anchors.
	 *
	 * @return the {$link URL} of this module descriptor.
	 */
	public URL getBase() {
		return base;
	}

	/**
	 * Retrieves the name of this module.
	 *
	 * @return the name of this module as specified in the descriptor.
	 */
	public String getName() {
		return editor.getValue( "module:name/text()");
	}

	/**
	 * Retrieves the version of this module.
	 *
	 * @return the version of this module as specified in the descriptor.
	 */
	public String getVersion() {
		return editor.getValue( "module:version/text()");
	}

	/**
	 * Retrieves a {$link String} containing the path to the <a href='http://java.sun.com/j2se/javadoc/'>Javadoc</a>
	 * documentation for the code this module packages.  This setting is optional.
	 *
	 * @return the path to this module's documentation or the empty {$link String} if none exists.
	 */
	public String getJavadoc() {
		return editor.getValue( "module:javadoc/text()");
	}

	/**
	 * Returns a node containing the content of the extensions section of the
	 * descriptor for this module. Concrete extension classes should filter the
	 * resulting node by namespace to retrieve relevant extension code.
	 * <p/>
	 * TODO: The content should be accessible by namespace and non-prefixed path
	 * only.
	 *
	 * @param namespace the namespace of the extension.
	 * @param prefix the prefix within the document of this extension.
	 * @param path the prefixed path to the setting relative to the extensions element.
	 * @return a {@link String} representing the extension setting at the given path.
	 */
	public String getExtension( final String namespace, final String prefix, final String path) {
		editor.startPrefixMapping( namespace, prefix);
		return editor.getValue( "module:extensions/" + path);
	}

	/**
	 * Returns a node containing the content of the extensions section of the
	 * descriptor for this module. Concrete extension classes should filter the
	 * resulting node by namespace to retrieve relevant extension code.
	 *
	 * @return a {@link Element} containing the extensions element.
	 */
	public Element getExtensions() {
		return ( Element) editor.findNode( "module:extensions");
	}

	/**
	 * Loads a component {@link URL} by alias.  If this module has been resolved via the {@link #resolve()} method, this
	 * call will fall through to component modules in the event that this module has no associated alias.
	 *
	 * @param alias component alias to dereference.
	 * @return a component {@link URL} corresponding to the alias.
	 */
	public URL getComponent( final String alias) throws MalformedURLException {
		final String uri = editor.getValue( "module:component[ module:alias/text() = '" + alias + "']/module:uri/text()");
		if ( uri != null) return new URL( base, uri, null);
		if ( resolved == null) return null;

		for ( Module module : resolved) {
			final URL url = module.getComponent( alias);
			if ( url != null) {
				notifier.debug( "Module ''{0}'' found component ''{0}'' within indexed module ''{0}''.", this, alias, module);
				return url;
			}
		}
		return null;
	}

	/**
	 * Returns an {@link Iterable} collection of {@link URL} instances that refer to each
	 * of the components of this module.
	 *
	 * @throws MalformedURLException if any of the module components is an invalid reference.
	 */
	public Collection< URL> getComponents() throws MalformedURLException {
		return getUrls( "module:component/module:uri");
	}

	/**
	 * Returns an {@link Iterable} collection of {@link URL} instances that refer to each
	 * of the dependencies of this module.
	 *
	 * @throws MalformedURLException if any of the module dependencies is an invalid reference.
	 */
	public Collection< URL> getDependencies() throws MalformedURLException {
		return getUrls( "module:dependency");
	}

	/**
	 * Resolves component and dependent modules to enable fallthrough on alias lookups.  This
	 * works by locally caching and loading each component module definition.
	 *
	 * @throws MalformedURLException if any of the module components or dependencies is an invalid reference.
	 * @throws IOException if an error occurs while caching the artifact.
	 */
	public synchronized void resolve() throws MalformedURLException, IOException {
		if ( resolved != null) {
			notifier.debug( "Module ''{0}'' has been resolved, ignoring.", this);
			return;
		}
   		resolved = new LinkedList< Module>();

		notifier.debug( "Resolving dependencies for module ''{0}''.", this);
		for ( URL url : getComponents()) {
			final Object object = load( url);
			if ( object instanceof Module) {
				notifier.debug( "Indexing module ''{0}''.", object);
				(( Module) object).resolve();
				resolved.add(( Module) object);
			}
		}
	}

	protected Collection< URL> getUrls( final String path) throws MalformedURLException {
		final List< URL> list = new LinkedList< URL>();
		for ( Node node : editor.findNodes( path)) {
			try {
				editor.pushContext( node);
				list.add( new URL( base, editor.getValue( "text()"), null));
			} finally {
				editor.popContext();
			}
		}
		return list;
	}

	@Override
	public String toString() {
		return new StringBuilder().append( "Module[ ").append( getName()).append( ", ").append( getVersion()).append( "]").toString();
	}

	/**
	 * Determines the proper alias name for a given fully qualified versioned artifact.
	 *
	 * @return a normalized alias for the given artifact.
	 */
	public static String alias( final String uri) {
		return uri.replaceAll( EXTRACT_NAME[ 0], EXTRACT_NAME[ 1]);
	}

	/**
	 * Loads a module {@link URL} and any fragment part it may contain.  If this {@link URL} references
	 * a module definition it will be loaded and, if it has no fragment, returned as is.  If a fragment is
	 * present it will be resolved to a {@link URI} and returned.  If the provided {@link URL} does not
	 * reference a module definition, <code>null</code> is returned.
	 *
	 * @param url {@link URL} describing the module definition to load.
	 * @return an instance of {@link Module} if the {@link URL} references a module, a {@link URL} if it has a fragment that
	 * points to a particular component of the module, or <code>null</code> if it does not reference a module.
	 */
	public static Object load( final URL url) throws MalformedURLException, IOException {
		final String string = url.toString();
		notifier.debug( "Attempting to load url ''{0}''.", string);
		if ( cache.containsKey( string)) {
			notifier.debug( "Using cached modules for ''{0}''.", string);
			return cache.get( string);
		}

		try {
			if ( isBrimstoneUri( url)) {
				notifier.debug( "Caching Brimstone uri ''{0}''.", url);
				loadToCache( url.toURI());
			}
		} catch ( URISyntaxException e) {
			notifier.debug( "Loading uri ''{0}'' failed with an exception.", e);
		}
		
		final Object object = url.getContent( new Class[] { Module.class});
		if ( object instanceof Module) {
			notifier.debug( "Creating new module for ''{0}''.", object);
			final Module module = ( Module) object;
			final String fragment = url.getRef();
			if ( fragment != null) {
				module.resolve();
				return module.getComponent( fragment);
			} else {
				cache.put( string, module);
				return module;
			}
		}
		return null;
	}
}
