package org.freecompany.brimstone.module;

import org.freecompany.brimstone.cache.Util;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.Map;
import java.util.HashMap;
import java.util.logging.*;

import static java.util.logging.Level.FINE;
import static java.util.logging.Level.FINER;

/**
 * Extends the standard {@link URLClassLoader} implementation to expose functions
 * for adding {@link URL} instances to the base {@link URLClassLoader} at runtime. It
 * also supports adding a set of class collections to the loader using a Brimstone {@link Module}.
 */
public class ModuleClassLoader extends DelegatingClassLoader {

	protected static Logger logger = Logger.getLogger( ModuleClassLoader.class.getName());

	public ModuleClassLoader( final URL url) throws IllegalArgumentException {
		this( url, null);
	}

	public ModuleClassLoader( final URL url, final ClassLoader parent) throws IllegalArgumentException {
		super( parent);
		addUrl( url);
	}

	public ModuleClassLoader( final URL[] urls) throws IllegalArgumentException {
		this( urls, null);
	}

	public ModuleClassLoader( final URL[] urls, final ClassLoader parent) throws IllegalArgumentException {
		super( parent);
		for ( URL url : urls) addUrl( url);
	}

	public ModuleClassLoader( final Module module) throws IllegalArgumentException {
		this( module, null);
	}

	public ModuleClassLoader( final Module module, final ClassLoader parent) throws IllegalArgumentException {
		super( parent);
		addModule( module);
	}

	/**
	 * Extends the default add method to support automatic resolution of module sources. If
	 * a {@link URL} provided to this method provides the location of a module descriptor, it
	 * will be loaded and encapsulated in a new class loader as a child of this one.
	 *
	 * @param url the url to add to list of locations searched for classes.
	 */
	@Override
	public void addUrl( final URL url) {
		try {
			final Object object = Module.load( url);
			if ( object instanceof Module) {
				logger.log( FINE, "Adding module ''{0}''.", url);
				addDelegate( new ModuleClassLoader(( Module) object, this));
			} else if ( object instanceof URL) {
				logger.log( FINE, "Adding fragment ''{0}''.", url);
				addUrl(( URL) object);
			} else {
				logger.log( FINE, "Adding url ''{0}''.", url);
				super.addUrl( url);
			}
		} catch ( IOException e) {
			throw new IllegalArgumentException( "Error occured while loading resource '" + url + "'.", e);
		}
	}

	/**
	 * Adds a module definition to this class loader.  This method adds all of the dependencies and
	 * components of the provided module into this class loader.
	 * 
	 * TODO: This should behave like the addURL call and simply add a new ModuleClassLoader as a child.
	 * 
	 * @param module the module to add to this class loader.
	 */
	public void addModule( final Module module) {
		logger.log( FINE, "Delegating to module ''{0}''.", module);
		try {
			module.resolve();
			for ( URL url : module.getComponents()) addUrl( url);
			for ( URL url : module.getDependencies()) addUrl( url);
		} catch ( MalformedURLException e) {
			throw new IllegalArgumentException( "The provided module file has invalid entries.", e);
		} catch ( IOException e) {
			throw new IllegalArgumentException( "Error resolving linked modules.", e);
		}
	}
}
